unit Graphics;

interface

uses Defs;

type
  hBitmap=record
    data: Pointer;
    w, h: Word;
    pal:  array[1..1024] of Byte;
  end;

procedure SetMode(mode: Word);
procedure PutPixel(x, y: Integer; col: Byte);
function  GetPixel(x, y: Integer): Byte;
procedure Cls(col: Byte);
function  GetMode: Byte;
procedure Bar(x1, y1, x2, y2: Integer; col: Byte);
procedure Bar_(x, y, w, h: Integer; col: Byte);
procedure Rect(x1, y1, x2, y2: Integer; col: Byte);
procedure Rect_(x, y, w, h: Integer; col: Byte);
procedure GetImage(x, y, w, h: Integer; buf: Pointer);
procedure PutImage(x1, y1, w, h, x2, y2, w_: Integer; buf: Pointer);
procedure PutImage_(x1, y1, w, h, x2, y2, w_: Integer; buf: Pointer);
procedure PutBitmap(x, y: Integer; b: hBitmap);
procedure Flip_V(b: hBitmap);
procedure Flip_H(b: hBitmap);
procedure Flip__H(w, h, Offs: Integer; Data: Pointer);
procedure VertLine(x, y1, y2: Integer; col: Byte);
procedure VertLine_(x, y, h: Integer; col: Byte);
procedure HorLine(x1, y, x2: Integer; col: Byte);
procedure HorLine_(x, y, w: Integer; col: Byte);

procedure ScrollUp(lines: Integer);
procedure ScrollDown(lines: Integer);
procedure ScrollLeft(lines: Integer);
procedure ScrollRight(lines: Integer);
procedure ScrollXY(x, y: Integer);

function  SetBuf: Pointer;
procedure FreeBuf(buf: Pointer);
procedure CopyToScreen(buf: Pointer);
procedure CopyFromScreen(buf: Pointer);
procedure CopyToBuf(buf1, buf2: Pointer);
procedure SetScreenBuf(buf: Pointer);
procedure RestoreScreenBuf;

procedure SaveBmp(b: hBitmap; filename: String);
procedure LoadBmp(var b: hBitmap; filename: String);

procedure SetRGB(n, r, g, b: Byte);
procedure SetBmpPal(b: hBitmap);
procedure VertRet;

procedure OutChar(x, y: Integer; C: Char; Col: Byte);
procedure OutString(x, y: Integer; S: String; Col: Byte);

var ScreenBuf, Font: Pointer;
    FontSize: Byte;

implementation

procedure SetMode(mode: Word); assembler;
asm
    mov ax, mode
    int 10h
end;

procedure PutPixel(x, y: Integer; col: Byte); assembler;
asm
    cmp x, 319
    ja @1
    cmp y, 199
    ja @1
    les di, ScreenBuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov al, col
    stosb
@1:
end;

function  GetPixel(x, y: Integer): Byte;
  var col: Byte;
begin
  asm
    cmp x, 319
    ja @1
    cmp y, 199
    ja @1
    les di, ScreenBuf
    mov ax, x
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov al, es:[di]
    mov col, al
@1:
  end;
  GetPixel:=col;
end;

procedure Cls(col: Byte); assembler;
asm
  les di, ScreenBuf
  mov cx, 64000
  mov al, col
  rep stosb
end;

function  GetMode: Byte;
  var mode: byte;
begin
  asm
    mov ah, 0Fh
    int 10h
    mov mode, al
  end;
  GetMode:=mode;
end;

procedure Bar(x1, y1, x2, y2: Integer; col: Byte);
  var w, h: Integer;
begin
  if x2>x1 then begin w:=x2; x2:=x1; x1:=w; end;
  if y2>y1 then begin h:=y2; y2:=y1; y1:=h; end;

  if x1<0 then x1:=0;
  if y1<0 then y1:=0;

  if x2>320 then x2:=320;
  if y2>200 then y2:=200;

  w:=x2-x1;
  h:=y2-y1;

  if (x1<320) and (y1<200) and (x2>=0) and (y2>=0) then
  asm
    les di, ScreenBuf
    mov ax, y1
    mov bx, 320
    mul bx
    add ax, x1
    add di, ax
    mov al, col
    mov cx, h
@1: push cx
    mov cx, w
    rep stosb
    sub di, w
    add di, 320
    pop cx
    loop @1
  end;
end;

procedure Bar_(x, y, w, h: Integer; col: Byte);
begin
  if x<0 then begin w:=w+x; x:=0; end;
  if y<0 then begin h:=h+y; y:=0; end;

  if x+w>320 then w:=320-x;
  if y+h>200 then h:=200-y;

  if (w>0) and (h>0) then
  asm
    cmp x, 319
    ja @2
    cmp y, 199
    ja @2
    les di, ScreenBuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov al, col
    mov cx, h
@1: push cx
    mov cx, w
    rep stosb
    sub di, w
    add di, 320
    pop cx
    loop @1
@2:
  end;
end;

procedure Rect(x1, y1, x2, y2: Integer; col: Byte);
begin
  HorLine(x1, y1, x2, col);
  HorLine(x1, y2, x2, col);
  VertLine(x1, y1, y2, col);
  VertLine(x2, y1, y2+1, col);
end;

procedure Rect_(x, y, w, h: Integer; col: Byte);
begin
  HorLine_(x, y, w, col);
  HorLine_(x, y+h, w, col);
  VertLine_(x, y, h, col);
  VertLine_(x+w, y, h+1, col);
end;

procedure GetImage(x, y, w, h: Integer; buf: Pointer);
begin
  if (x>=0) and (x+w<321) and (y>=0) and (y+h<201) then
  asm
    push ds
    lds si, ScreenBuf
    les di, buf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add si, ax
    mov cx, h
@1: push cx
    mov cx, w
    rep movsb
    sub di, w
    add di, 320
    pop cx
    loop @1
    pop ds
  end;
end;

procedure PutImage(x1, y1, w, h, x2, y2, w_: Integer; buf: Pointer);
  var offs: Integer;
begin
  offs:=y2*w_+x2;

  if y1<0 then begin offs:=offs+w*(-y1); h:=h+y1; y1:=0; end;
  if x1<0 then begin offs:=offs-x1; w:=w+x1; x1:=0; end;

  if x1+w>320 then w:=320-x1;
  if y1+h>200 then h:=200-y1;

  if (w>0) and (h>0) then
  asm
    cmp x1, 319
    ja @2
    cmp y1, 199
    ja @2
    push ds
    les di, ScreenBuf
    lds si, buf
    mov ax, y1
    mov bx, 320
    mul bx
    add ax, x1
    add si, offs
    add di, ax
    mov cx, h
@1: push cx
    mov cx, w
@3: mov al, ds:[si]
    cmp al, RENDER_COLOR
    je @4
    movsb
    jmp @5
@4: inc si
    inc di
@5: loop @3
    sub di, w
    add di, 320
    sub si, w
    add si, w_
    pop cx
    loop @1
    pop ds
@2:
  end;
end;

procedure PutImage_(x1, y1, w, h, x2, y2, w_: Integer; buf: Pointer);
  var offs: Integer;
begin
  offs:=y2*w_+x2;

  if y1<0 then begin offs:=offs+w*(-y1); h:=h+y1; y1:=0; end;
  if x1<0 then begin offs:=offs-x1; w:=w+x1; x1:=0; end;

  if x1+w>320 then w:=320-x1;
  if y1+h>200 then h:=200-y1;

  if (w>0) and (h>0) then
  asm
    cmp x1, 319
    ja @2
    cmp y1, 199
    ja @2
    push ds
    les di, ScreenBuf
    lds si, buf
    mov ax, y1
    mov bx, 320
    mul bx
    add ax, x1
    add si, offs
    add di, ax
    mov cx, h
@1: push cx
    mov cx, w
    rep movsb
    sub di, w
    add di, 320
    sub si, w
    add si, w_
    pop cx
    loop @1
    pop ds
@2:
  end;
end;

procedure PutBitmap(x, y: Integer; b: hBitmap);
begin
  PutImage(x, y, b.w, b.h, 0, 0, b.w, b.data);
end;

procedure Flip_V(b: hBitmap);
  var buf: Pointer;
      n: Integer;
begin
  n:=b.w*b.h;
  GetMem(buf, n);

  asm
    push ds
    les di, buf
    lds si, b.data
    mov cx, n
    rep movsb
    pop ds
  end;

  asm
    push ds
    les di, b.data
    lds si, buf
    mov cx, b.h
    dec cx
@1: push cx
    mov ax, cx
    dec ax
    mov bx, b.w
    mul bx
    mov si, ax
    mov cx, b.w
    rep movsb
    pop cx
    loop @1
    pop ds
  end;

  FreeMem(buf, n);
end;

procedure Flip_H(b: hBitmap);
  var buf: Pointer;
      n: Integer;
begin
  n:=b.w*b.h;
  GetMem(buf, n);

  asm
    push ds
    les di, buf
    lds si, b.data
    mov cx, n
    rep movsb
    pop ds
  end;

  asm
    push ds
    les di, b.data
    lds si, buf
    mov cx, b.h
    add di, b.w
    sub di, 2
@1: push cx
    mov cx, b.w
@2: movsb
    sub di, 2
    loop @2
    pop cx
    mov ax, b.w
    shl ax, 1
    add di, ax
    loop @1
    pop ds
  end;

  FreeMem(buf, n);
end;

procedure Flip__H(w, h, Offs: Integer; Data: Pointer);
  var buf: Pointer;
      n: Integer;
begin
  n := w * h;
  GetMem(buf, n);

  asm
    push ds
    les di, buf
    lds si, Data
    add si, Offs
    mov cx, n
    rep movsb
    pop ds
  end;

  asm
    push ds
    les di, Data
    add di, Offs
    lds si, buf
    mov cx, h
    add di, w
    dec di
@1: push cx
    mov cx, w
@2: movsb
    sub di, 2
    loop @2
    pop cx
    mov ax, w
    shl ax, 1
    add di, ax
    loop @1
    pop ds
  end;

  FreeMem(buf, n);
end;

procedure VertLine(x, y1, y2: Integer; col: Byte);
  var h: Integer;
begin
  if y1>y2 then begin h:=y1; y1:=y2; y2:=h; end;

  if y1<0 then y1:=0;
  if y2>199 then y2:=199;

  h:=y2-y1;

  if (y1<200) and (y2>=0) and (h>0) then
  asm
    les di, ScreenBuf
    mov ax, y1
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov cx, h
    mov al, col
@1: stosb
    add di, 319
    loop @1
  end;
end;

procedure VertLine_(x, y, h: Integer; col: Byte);
begin
  if y<0 then begin h:=h+y; y:=0; end;

  if h>0 then
  asm
    cmp x, 319
    ja @2
    cmp y, 199
    ja @2
    les di, ScreenBuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov cx, h
    mov al, col
@1: stosb
    add di, 319
    loop @1
@2:
  end;
end;

procedure HorLine(x1, y, x2: Integer; col: Byte);
  var w: Integer;
begin
  if x1>x2 then begin w:=x1; x1:=x2; x2:=w; end;

  if x1<0 then x1:=0;
  if x2>319 then x2:=319;

  w:=x2-x1;

  if (x1<320) and (x2>=0) and (w>0) then
  asm
    les di, ScreenBuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x1
    add di, ax
    mov cx, w
    mov al, col
    rep stosb
  end;
end;

procedure HorLine_(x, y, w: Integer; col: Byte);
begin
  if x<0 then begin w:=w+x; x:=0; end;

  if w>0 then
  asm
    cmp x, 319
    ja @2
    cmp y, 199
    ja @2
    les di, ScreenBuf
    mov ax, y
    mov bx, 320
    mul bx
    add ax, x
    add di, ax
    mov cx, w
    mov al, col
    rep stosb
@2:
  end;
end;

procedure ScrollXY(x, y: Integer);
  var buf: Pointer;
begin
  GetMem(buf, 64000);
  GetImage(0, 0, 320, 200, buf);
  PutImage(x, y, 320, 200, 0, 0, 320, buf);
  FreeMem(buf, 64000);
end;

procedure ScrollUp(lines: Integer);
begin
  ScrollXY(0, lines);
end;

procedure ScrollDown(lines: Integer);
begin
  ScrollXY(0, -lines);
end;

procedure ScrollRight(lines: Integer);
begin
  ScrollXY(-lines, 0);
end;

procedure ScrollLeft(lines: Integer);
begin
  ScrollXY(lines, 0);
end;

function  SetBuf: Pointer;
  var p: Pointer;
begin
  GetMem(p, 64000);
  SetBuf:=p;
end;

procedure FreeBuf(buf: Pointer);
begin
  FreeMem(buf, 64000);
end;

procedure CopyToScreen(buf: Pointer); assembler;
asm
    push ds
    lds si, buf
    mov ax, 0A000h
    mov es, ax
    xor di, di
    mov cx, 32000
    rep movsw
    pop ds
end;

procedure CopyFromScreen(buf: Pointer); assembler;
asm
    push ds
    les di, buf
    mov ax, 0A000h
    mov ds, ax
    xor si, si
    mov cx, 32000
    rep movsw
    pop ds
end;

procedure CopyToBuf(buf1, buf2: Pointer); assembler;
asm
    push ds
    les di, buf2
    lds si, buf1
    mov cx, 32000
    rep movsw
    pop ds
end;

procedure SetScreenBuf(buf: Pointer);
begin
  ScreenBuf:=buf;
end;

procedure RestoreScreenBuf;
begin
  ScreenBuf:=Ptr($A000, 0);
end;

procedure SaveBmp(b: hBitmap; filename: String);
  var f: File;
begin
  Assign(f, filename);
  Rewrite(f);

  BlockWrite(f, b.pal, 1024);
  BlockWrite(f, b.data^, b.w*b.h);

  Close(f);
end;

procedure LoadBmp(var b: hBitmap; filename: String);
  var buf: Pointer;
      f: File;
begin
  Assign(f, filename);
  Reset(f, 1);

  Seek(f, 18);
  BlockRead(f, b.w, 2);
  Seek(f, 22);
  BlockRead(f, b.h, 2);

  Seek(f, 55);
  BlockRead(f, b.pal, 1024);

  GetMem(buf, b.w*b.h);
  BlockRead(f, buf^, b.w*b.h);
  b.data:=buf;

  Close(f);
end;

procedure SetRGB(n, r, g, b: Byte); assembler;
asm
  mov dx, 3C8h
  mov al, n
  out dx, al
  inc dx
  mov al, r
  shr al, 2
  out dx, al
  mov al, g
  shr al, 2
  out dx, al
  mov al, b
  shr al, 2
  out dx, al
end;

procedure SetBmpPal(b: hBitmap);
  var i: Integer;
begin
  for i:=0 to 255 do
    SetRGB(i, b.pal[i*4+2], b.pal[i*4+1], b.pal[i*4]);
end;

procedure VertRet; assembler;
asm
{    cli}
    mov dx, 03DAh
@1: in al, dx
    test al, 8
    jz @1
@2: in al, dx
    test al, 8
    jnz @2
{    sti  }
end;

procedure OutChar(x, y: Integer; C: Char; Col: Byte);
  var i, j: Integer;
      B: Byte;
begin

  for i := 0 to FontSize - 1 do begin
    asm
      push es
      les di, Font
      mov ah, 0
      mov al, C
      mov bl, FontSize
      mul bl
      add ax, i
      add di, ax
      mov al, es:[di]
      mov B, al
      pop es
    end;
    for j := 0 to 7 do begin
      if B and (1 shl j) <> 0 then
        PutPixel(x + (7 - j), y + i, Col);
    end;
  end;
end;

procedure OutString(x, y: Integer; S: String; Col: Byte);
  var i: Integer;
begin
  for i := 1 to length(S) do
    OutChar(x + (i - 1) * 9, y, S[i], Col);
end;

end.