unit Monsters;

interface

uses Defs, Procs, Graphics;

type
    PMonster = ^TMonster;
    TMonster = object
      Dir, Dir_, x, y, MonsterType, FlipFrame: Integer;
      Frame, FrameTime, Frames, Size, StartFrame: Integer;
      Next: PMonster;
      NeedFlip: Boolean;
      constructor Init(PosX, PosY, Morph, Direction: Integer);
      procedure MoveMonster(dx, dy: Integer);
      procedure HandleEvent;
      procedure Draw(ScreenX, ScreenY: Integer);
      destructor Done;
    end;

implementation

constructor TMonster.Init(PosX, PosY, Morph, Direction: Integer);
begin
  x := PosX;
  if Morph = 1 then
    NeedFlip := False
  else
    NeedFlip := True;
  if Morph = 0 then
    Size := 2
  else
    Size := 1;
  if Morph > 2 then
    Frames := 0
  else
    Frames := 1;
  MonsterType := Morph;
  StartFrame := MONSTER_FRAMES + FRAME_TABLE[Morph + 1];
  FrameTime := MONSTER_FRAME_TIME;
  Frame := 0;
  Dir := Direction;
  Dir_ := Dir;
  FlipFrame := 0;
  if (Dir <> -1) and NeedFlip then
    FlipFrame := 1;
  if Morph <> 3 then
    y := PosY
  else begin
    Dir_ := PosY;
    y := LEVEL_HEIGHT + 14;
    Dir := -1;
    FrameTime := FISH_PAUSE_TIME;
  end;
  Next := nil;
end;

procedure TMonster.HandleEvent;
begin
  if MonsterType <> 3 then begin
    MoveMonster(Dir * MONSTER_SPEED, 0);
    if (Dir <> Dir_) and NeedFlip then
      if FlipFrame = 0 then
        FlipFrame := 1
      else
        FlipFrame := 0;
    Dir_ := Dir;
    MoveMonster(0, MONSTER_DOWN_SPEED);
    Dec(FrameTime);
    if FrameTime = 0 then begin
      if Frame = 0 then
        Frame := Frames
      else
        Frame := 0;
      FrameTime := MONSTER_FRAME_TIME;
    end;
  end
  else begin
    MoveMonster(0, Dir * MONSTER_DOWN_SPEED);
    if y <= Dir_ then Dir := 1;
    if y > LEVEL_HEIGHT + 14 then begin
      Dir := 0;
      Dec(FrameTime);
      if FrameTime = 0 then begin
        Dir := -1;
        FrameTime := FISH_PAUSE_TIME;
      end;
    end;
  end;
  if Next <> nil then Next^.HandleEvent;
end;

procedure TMonster.Draw(ScreenX, ScreenY: Integer);
begin
  PutImage(x - ScreenX, y - ScreenY - (14 * (Size - 1)), 20, Size * 14, 0,
  (StartFrame + (Frame * Size) + (FlipFrame * Size * Frames * 2)) * 14, 20, Data);
  if Next <> nil then Next^.Draw(ScreenX, ScreenY);
end;

procedure TMonster.MoveMonster(dx, dy: Integer);
  var Allow: Boolean;
      TempX, TempY, i, j: Integer;
      n: Byte;
begin
  TempX := x + dx;
  TempY := y + dy;
  Allow := True;

  for i := (TempX - dx) div 20 - 1 to (TempX - dx) div 20 + 1 do
  for j := (TempY - dy) div 14 - 1 to (TempY - dy) div 14 + 1 do begin
    if (TempX > i * 20 - 20) and (TempX < i * 20 + 20) then
    if (TempY > j * 14 - 14) and (TempY < j * 14 + Size * 14) then begin
      n := Level^[i + 1, j + 1];
      if Allow then
        case SOLID_TABLE[n + 1] of
          0: Allow := True;
          1, 2: Allow := False;
        end;
    end;
  end;

  if (x < 0) or (x > LEVEL_WIDTH - 20) then Allow := False;

  if Allow then begin
    x := x + dx;
    y := y + dy;
  end
  else if (dx <> 0) and (MonsterType <> 3) then begin
    if Dir = 1 then
      Dir := -1
    else
      Dir := 1;
  end;
end;

destructor TMonster.Done;
begin
  if Next <> nil then Dispose(Next, Done);
end;

end.
